/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// d_sky.c

#include "quakedef.h"
#include "r_local.h"
#include "d_local.h"

#define SKY_SPAN_SHIFT	5
#define SKY_SPAN_MAX	(1 << SKY_SPAN_SHIFT)

// Manoel Kasimier - begin
extern byte
	*skyunderlay
,	*skyoverlay
	;

static byte
	*pdest
,	over
	;
static float
	timespeed1
,	timespeed2
	;
static espan_t
	*pspan
	;
static int
	count
,	spancount
,	spancountminus1
,	u
,	v
,	stipple
	;
static fixed16_t
	s
,	t
,	snext
,	tnext
,	sstep
,	tstep
,	s2
,	t2
,	snext2
,	tnext2
,	sstep2
,	tstep2
	;
// Manoel Kasimier - end


/*
=================
D_Sky_uv_To_st
=================
*/
void D_Sky_uv_To_st (int u, int v, fixed16_t *s, fixed16_t *t
					, fixed16_t *s2, fixed16_t *t2) // Manoel Kasimier - smooth sky
{
	float	wu, wv;//, temp; // Manoel Kasimier - edited
	vec3_t	end;

	// ToChriS - begin
	wu = (u - xcenter) / xscale;
	wv = (ycenter - v) / yscale;

	end[0] = vpn[0] + wu*vright[0] + wv*vup[0];
	end[1] = vpn[1] + wu*vright[1] + wv*vup[1];
	end[2] = vpn[2] + wu*vright[2] + wv*vup[2];
	// ToChriS - end
	end[2] *= 3;
	VectorNormalize (end);
	*s = (int)((timespeed1 + 6*(SKYSIZE/2-1)*end[0]) * 0x10000); // Manoel Kasimier - smooth sky - edited
	*t = (int)((timespeed1 + 6*(SKYSIZE/2-1)*end[1]) * 0x10000); // Manoel Kasimier - smooth sky - edited
	*s2 = (int)((timespeed2 + 6*(SKYSIZE/2-1)*end[0]) * 0x10000); // Manoel Kasimier - smooth sky
	*t2 = (int)((timespeed2 + 6*(SKYSIZE/2-1)*end[1]) * 0x10000); // Manoel Kasimier - smooth sky
}


/*
=================
D_DrawSkyScans8
=================
*/
void D_DrawSkyScans_Transparent ()
{
	do
	{
		pdest = (byte *) ( (byte *) d_viewbuffer + (screenwidth * pspan->v) + pspan->u);
		count = pspan->count;

		// calculate the initial s & t
		u = pspan->u;
		v = pspan->v;
		D_Sky_uv_To_st (u, v, &s, &t, &s2, &t2); // Manoel Kasimier - smooth sky - edited

		do
		{
			spancount = (count >= SKY_SPAN_MAX) ? SKY_SPAN_MAX : count; // Manoel Kasimier

			count -= spancount;

			if (count)
			{
				u += spancount;

			// calculate s and t at far end of span,
			// calculate s and t steps across span by shifting
				D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

				sstep = (snext - s) >> SKY_SPAN_SHIFT;
				tstep = (tnext - t) >> SKY_SPAN_SHIFT;
			}
			else
			{
			// calculate s and t at last pixel in span,
			// calculate s and t steps across span by division
				spancountminus1 = (float)(spancount - 1);

				if (spancountminus1 > 0)
				{
					u += spancountminus1;
					D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

					sstep = (snext - s) / spancountminus1;
					tstep = (tnext - t) / spancountminus1;
				}
			}

			do
			{
				*pdest++ = skyunderlay[((t  & R_SKY_TMASK) >> 8) + ((s  & R_SKY_SMASK) >> 16)]; // Manoel Kasimier
				s += sstep;
				t += tstep;
			} while (--spancount > 0);

			s = snext;
			t = tnext;

		} while (count > 0);

	} while ( (pspan = pspan->pnext) != NULL);
}

void D_DrawSkyScans_Solid ()
{
	do
	{
		pdest = (byte *) ( (byte *) d_viewbuffer + (screenwidth * pspan->v) + pspan->u);
		count = pspan->count;

		// calculate the initial s & t
		u = pspan->u;
		v = pspan->v;
		D_Sky_uv_To_st (u, v, &s, &t, &s2, &t2); // Manoel Kasimier - smooth sky - edited

		do
		{
			spancount = (count >= SKY_SPAN_MAX) ? SKY_SPAN_MAX : count; // Manoel Kasimier

			count -= spancount;

			if (count)
			{
				u += spancount;

			// calculate s and t at far end of span,
			// calculate s and t steps across span by shifting
				D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

				sstep = (snext - s) >> SKY_SPAN_SHIFT;
				tstep = (tnext - t) >> SKY_SPAN_SHIFT;
				sstep2 = (snext2 - s2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
				tstep2 = (tnext2 - t2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
			}
			else
			{
			// calculate s and t at last pixel in span,
			// calculate s and t steps across span by division
				spancountminus1 = (float)(spancount - 1);

				if (spancountminus1 > 0)
				{
					u += spancountminus1;
					D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

					sstep = (snext - s) / spancountminus1;
					tstep = (tnext - t) / spancountminus1;
					sstep2 = (snext2 - s2) / spancountminus1; // Manoel Kasimier - smooth sky
					tstep2 = (tnext2 - t2) / spancountminus1; // Manoel Kasimier - smooth sky
				}
			}

			do
			{
				*pdest++ = ( (over = skyoverlay [ ( (t2 & R_SKY_TMASK) >> 8) + ( (s2 & R_SKY_SMASK) >> 16)]) == 255) ? skyunderlay[((t  & R_SKY_TMASK) >> 8) + ((s  & R_SKY_SMASK) >> 16)] : over; // Manoel Kasimier - smooth sky
				s += sstep;
				t += tstep;
				s2 += sstep2; // Manoel Kasimier - smooth sky
				t2 += tstep2; // Manoel Kasimier - smooth sky
			} while (--spancount > 0);

			s = snext;
			t = tnext;
			s2 = snext2; // Manoel Kasimier - smooth sky
			t2 = tnext2; // Manoel Kasimier - smooth sky

		} while (count > 0);

	} while ( (pspan = pspan->pnext) != NULL);
}

void D_DrawSkyScans_Stipple ()
{
	do
	{
		pdest = (byte *) ( (byte *) d_viewbuffer + (screenwidth * pspan->v) + pspan->u);
		stipple = ( ( (long) pdest - ( (long) pdest - (long) d_viewbuffer) / (long) screenwidth)) & 1; // Manoel Kasimier - stippled sky
		count = pspan->count;

		// calculate the initial s & t
		u = pspan->u;
		v = pspan->v;
		D_Sky_uv_To_st (u, v, &s, &t, &s2, &t2); // Manoel Kasimier - smooth sky - edited

		do
		{
			spancount = (count >= SKY_SPAN_MAX) ? SKY_SPAN_MAX : count; // Manoel Kasimier

			count -= spancount;

			if (count)
			{
				u += spancount;

			// calculate s and t at far end of span,
			// calculate s and t steps across span by shifting
				D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

				sstep = (snext - s) >> SKY_SPAN_SHIFT;
				tstep = (tnext - t) >> SKY_SPAN_SHIFT;
				sstep2 = (snext2 - s2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
				tstep2 = (tnext2 - t2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
			}
			else
			{
			// calculate s and t at last pixel in span,
			// calculate s and t steps across span by division
				spancountminus1 = (float)(spancount - 1);

				if (spancountminus1 > 0)
				{
					u += spancountminus1;
					D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

					sstep = (snext - s) / spancountminus1;
					tstep = (tnext - t) / spancountminus1;
					sstep2 = (snext2 - s2) / spancountminus1; // Manoel Kasimier - smooth sky
					tstep2 = (tnext2 - t2) / spancountminus1; // Manoel Kasimier - smooth sky
				}
			}

			do
			{
				// Manoel Kasimier - stippled sky - begin
				*pdest++ = (stipple = !stipple)
				?	skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)]
				:	( (over = skyoverlay [ ( (t2 & R_SKY_TMASK) >> 8) + ( (s2 & R_SKY_SMASK) >> 16)]) == 255) ? skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)] : over;
					;
				// Manoel Kasimier - stippled sky - end
				s += sstep;
				t += tstep;
				s2 += sstep2; // Manoel Kasimier - smooth sky
				t2 += tstep2; // Manoel Kasimier - smooth sky
			} while (--spancount > 0);

			s = snext;
			t = tnext;
			s2 = snext2; // Manoel Kasimier - smooth sky
			t2 = tnext2; // Manoel Kasimier - smooth sky
		} while (count > 0);

	} while ( (pspan = pspan->pnext) != NULL);
}

void D_DrawSkyScans_Blend ()
{
	colorblendingmap = colorblendingmode[minmaxclamp (0, (int) r_sky_blend_map.value, COLORBLENDING_MAPS - 1)];
	do
	{
		pdest = (byte *) ( (byte *) d_viewbuffer + (screenwidth * pspan->v) + pspan->u);
		count = pspan->count;

		// calculate the initial s & t
		u = pspan->u;
		v = pspan->v;
		D_Sky_uv_To_st (u, v, &s, &t, &s2, &t2); // Manoel Kasimier - smooth sky - edited

		do
		{
			spancount = (count >= SKY_SPAN_MAX) ? SKY_SPAN_MAX : count; // Manoel Kasimier

			count -= spancount;

			if (count)
			{
				u += spancount;

			// calculate s and t at far end of span,
			// calculate s and t steps across span by shifting
				D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

				sstep = (snext - s) >> SKY_SPAN_SHIFT;
				tstep = (tnext - t) >> SKY_SPAN_SHIFT;
				sstep2 = (snext2 - s2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
				tstep2 = (tnext2 - t2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
			}
			else
			{
			// calculate s and t at last pixel in span,
			// calculate s and t steps across span by division
				spancountminus1 = (float)(spancount - 1);

				if (spancountminus1 > 0)
				{
					u += spancountminus1;
					D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

					sstep = (snext - s) / spancountminus1;
					tstep = (tnext - t) / spancountminus1;
					sstep2 = (snext2 - s2) / spancountminus1; // Manoel Kasimier - smooth sky
					tstep2 = (tnext2 - t2) / spancountminus1; // Manoel Kasimier - smooth sky
				}
			}

			do
			{
				over = skyoverlay [ ( (t2 & R_SKY_TMASK) >> 8) + ( (s2 & R_SKY_SMASK) >> 16)]; // Manoel Kasimier
				*pdest++ = (over == 255) ? skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)] : colorblendingmap[over + skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)] * 256]; // Manoel Kasimier - smooth sky
				s += sstep;
				t += tstep;
				s2 += sstep2; // Manoel Kasimier - smooth sky
				t2 += tstep2; // Manoel Kasimier - smooth sky
			} while (--spancount > 0);

			s = snext;
			t = tnext;
			s2 = snext2; // Manoel Kasimier - smooth sky
			t2 = tnext2; // Manoel Kasimier - smooth sky

		} while (count > 0);

	} while ( (pspan = pspan->pnext) != NULL);
}

void D_DrawSkyScans_BlendBackwards ()
{
	colorblendingmap = colorblendingmode[minmaxclamp (0, (int) r_sky_blend_map.value, COLORBLENDING_MAPS - 1)];
	do
	{
		pdest = (byte *) ( (byte *) d_viewbuffer + (screenwidth * pspan->v) + pspan->u);
		count = pspan->count;

		// calculate the initial s & t
		u = pspan->u;
		v = pspan->v;
		D_Sky_uv_To_st (u, v, &s, &t, &s2, &t2); // Manoel Kasimier - smooth sky - edited

		do
		{
			spancount = (count >= SKY_SPAN_MAX) ? SKY_SPAN_MAX : count; // Manoel Kasimier

			count -= spancount;

			if (count)
			{
				u += spancount;

			// calculate s and t at far end of span,
			// calculate s and t steps across span by shifting
				D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

				sstep = (snext - s) >> SKY_SPAN_SHIFT;
				tstep = (tnext - t) >> SKY_SPAN_SHIFT;
				sstep2 = (snext2 - s2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
				tstep2 = (tnext2 - t2) >> SKY_SPAN_SHIFT; // Manoel Kasimier - smooth sky
			}
			else
			{
			// calculate s and t at last pixel in span,
			// calculate s and t steps across span by division
				spancountminus1 = (float)(spancount - 1);

				if (spancountminus1 > 0)
				{
					u += spancountminus1;
					D_Sky_uv_To_st (u, v, &snext, &tnext, &snext2, &tnext2); // Manoel Kasimier - smooth sky - edited

					sstep = (snext - s) / spancountminus1;
					tstep = (tnext - t) / spancountminus1;
					sstep2 = (snext2 - s2) / spancountminus1; // Manoel Kasimier - smooth sky
					tstep2 = (tnext2 - t2) / spancountminus1; // Manoel Kasimier - smooth sky
				}
			}

			do
			{
				over = skyoverlay [ ( (t2 & R_SKY_TMASK) >> 8) + ( (s2 & R_SKY_SMASK) >> 16)]; // Manoel Kasimier
				*pdest++ = (over == 255) ? skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)] : colorblendingmap[skyunderlay[ ( (t  & R_SKY_TMASK) >> 8) + ( (s  & R_SKY_SMASK) >> 16)] + over * 256]; // Manoel Kasimier - smooth sky
				s += sstep;
				t += tstep;
				s2 += sstep2; // Manoel Kasimier - smooth sky
				t2 += tstep2; // Manoel Kasimier - smooth sky
			} while (--spancount > 0);

			s = snext;
			t = tnext;
			s2 = snext2; // Manoel Kasimier - smooth sky
			t2 = tnext2; // Manoel Kasimier - smooth sky

		} while (count > 0);

	} while ( (pspan = pspan->pnext) != NULL);
}

static void (*D_DrawSkyScans[2][COLORBLENDING_MAPS][4]) () = {
	{
		{	//BLEND_WEIGHTEDMEAN
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Solid
		}
	,	{	//BLEND_ADDITIVE
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Solid
		}
	,	{	//BLEND_SMOKE
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Stipple
		,	D_DrawSkyScans_Solid
		}
	}
,	{
		{	//BLEND_WEIGHTEDMEAN
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Blend
		,	D_DrawSkyScans_BlendBackwards
		,	D_DrawSkyScans_Solid
		}
	,	{	//BLEND_ADDITIVE
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Blend
		,	D_DrawSkyScans_BlendBackwards
		,	D_DrawSkyScans_BlendBackwards
		}
	,	{	//BLEND_SMOKE
			D_DrawSkyScans_Transparent
		,	D_DrawSkyScans_Blend
		,	D_DrawSkyScans_BlendBackwards
		,	D_DrawSkyScans_Solid
		}
	}
};

void D_DrawSkyScans8 (espan_t *span)
{
	// Manoel Kasimier - begin
	pspan = span;

	timespeed1 = skytime * skyspeed * r_skyspeed1.value;
	timespeed2 = skytime * skyspeed * r_skyspeed2.value;
	sstep2 = 0;
	tstep2 = 0;
	sstep = 0;	// keep compiler happy
	tstep = 0;	// ditto

	D_DrawSkyScans[minmaxclamp (0, (int) r_sky_blend_mode.value, 1)][minmaxclamp (0, (int) r_sky_blend_map.value, COLORBLENDING_MAPS - 1)][minmaxclamp (0, (int) (r_sky_blend_alpha.value * 3.0f), 3)] ();
	// Manoel Kasimier - end

}
